# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from arjuna.tpi.helper.image import Image
from arjuna.tpi.tracker import track

@track("debug")
class ScreenShooter(metaclass=abc.ABCMeta):

    def take_screenshot(self, *, prefix: str=None) -> Image:
        '''
            Take screenshot.

            Args:
                prefix: prefix to be appended to the screenshot name.

            Returns:
                `Image` object.
        '''
        img = self._take_screenshot(prefix=prefix)
        if not isinstance(img, Image):
            raise Exception("_take_screenshot for ScreenShooter protocol must return an Image object.")
        return img

    @abc.abstractclassmethod
    def _take_screenshot(self, *, prefix: str=None) -> Image:
        pass