.log {
    height: 500px;
}

.linebreaks {
    white-space:pre;            
}

div.image {
	border: 1px solid #e6e6e6;
	float: right;
    height: 500px;
	margin-left: 5px;
	overflow: scroll;
	width: 320px
}

body {font-family: Arial, Helvetica, sans-serif;}

/* The Modal (background) */
.modal {
  display: none; /* Hidden by default */
  position: fixed; /* Stay in place */
  z-index: 1; /* Sit on top */
  padding-top: 100px; /* Location of the box */
  left: 0;
  top: 0;
  width: 80%; /* Full width */
  height: 90%; /* Full height */
  overflow: auto;
  background-color: rgb(0,0,0); /* Fallback color */
  background-color: rgba(0,0,0,0.4); /* Black w/ opacity */
}

/* Modal Content */
.modal-content {
  background-color: #fefefe;
  margin: auto;
  padding: 5px;
  border: 1px solid #888;
  width: 80%;
  overflow: scroll;
}

.modal-content p{
    overflow: scroll;
}

/* The Close Button */
.close {
  color: #aaaaaa;
  float: right;
  font-size: 28px;
  font-weight: bold;
}

.close:hover,
.close:focus {
  color: #000;
  text-decoration: none;
  cursor: pointer;
}

.screenshot {
    border-radius: 5px;
    cursor: pointer;
    transition: 0.3s;
  }
  
.screenshot:hover {opacity: 0.7;}

button {
    margin-left: 10px;
    text-align: left;
}

button.redir_button {
    margin-left: 30px;
    text-align: left;
}

.noborder {
    border-collapse: collapse;
}

.noborder tbody, .noborder tr, .noborder td{
    text-align: left !important;
    border: 0 !important;
}

.noborder td{
    text-align: left !important;
}
